-- Turbine Imports..
import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

-- Plugin Imports..
import "CubePlugins.DeedTracker.Timer";
import "CubePlugins.DeedTracker.Globals";
import "CubePlugins.DeedTracker.PluginFunctions";
import "CubePlugins.DeedTracker.ChatLogger";
import "CubePlugins.DeedTracker.Commands";
import "CubePlugins.DeedTracker.GeneralFunctions";
import "CubePlugins.DeedTracker.Strings";
import "CubePlugins.DeedTracker.Onscreen";
import "CubePlugins.DeedTracker.VindarPatch";
import "CubePlugins.DeedTracker.DropDown";
import "CubePlugins.DeedTracker.Images";
import "CubePlugins.DeedTracker.DeedIDsToIndicies";
import "CubePlugins.DeedTracker.RegionInfo";
import "CubePlugins.DeedTracker.ConflictingQuests";
import "CubePlugins.DeedTracker.DeedItems";

-- Language Specific..
import "CubePlugins.DeedTracker.DeedInfoEN";
import "CubePlugins.DeedTracker.DeedInfoDE";
import "CubePlugins.DeedTracker.DeedInfoFR";
import "CubePlugins.DeedTracker.DeedNamesToIDsEN";
import "CubePlugins.DeedTracker.DeedNamesToIDsDE";
import "CubePlugins.DeedTracker.DeedNamesToIDsFR";
import "CubePlugins.DeedTracker.DeedObjectiveInfoEN";
import "CubePlugins.DeedTracker.DeedObjectiveInfoDE";
import "CubePlugins.DeedTracker.DeedObjectiveInfoFR";

-- Windows --
import "CubePlugins.DeedTracker.Options";
import "CubePlugins.DeedTracker.MainWin";
import "CubePlugins.DeedTracker.DeedTooltipWindow";
import "CubePlugins.DeedTracker.StringTooltipWindow";
import "CubePlugins.DeedTracker.DisambiguateWindow";
import "CubePlugins.DeedTracker.DeedOrQuestWindow";
import "CubePlugins.DeedTracker.CompletionWindow";
import "CubePlugins.DeedTracker.MinimizedIcon";

function UnloadReloader()
    Turbine.PluginManager.RefreshAvailablePlugins();
    local loadedPlugins = Turbine.PluginManager.GetLoadedPlugins();

    reloadedUnloader = Turbine.UI.Control();
    reloadedUnloader:SetWantsUpdates( true );
    
    reloadedUnloader.Update = function( sender, args )
        Turbine.PluginManager.RefreshAvailablePlugins();
        local loadedPlugins = Turbine.PluginManager.GetLoadedPlugins();

        for k,v in pairs(loadedPlugins) do
            if v.Name == "Deed Tracker Reloader" then
                Turbine.PluginManager.UnloadScriptState( 'DeedTrackerReloader' );
            end
        end
        reloadedUnloader:SetWantsUpdates( false );
    end
end

function GetCharDataFilename(characterName)
    local fixedName = characterName:gsub("-", "_");
    return "DeedTracker_CharData_" .. fixedName;
end

-- v1 format: All characters in one file, no separate :CHARACTERS entry in _CHARDATA
-- v2 format: All characters in separate files, :CHARACTERS entry in _CHARDATA tells us which files to look for
function saveData()
    saveCharacterSettings();
    saveCurrentCharacterData();
    saveServerSettings();

    if (SETTINGS.VERBOSE_OUTPUT) then
        Debug("  Deed Tracker: Save complete");
    end    
end

function saveCharacterSettings()
    PatchDataSave(Turbine.DataScope.Character, "DeedTracker_Settings", SETTINGS);
end

function saveServerSettings()
    local serverData = {};
    -- Write out special categories specially:
    for k,v in pairs (_CHARDATA) do
        if (k:sub(1, 1) == ":" and k ~= MYCHAR:GetName()) then -- Don't do special categories or the current charcater
            serverData[k] = deepcopy(v);
        end
    end
    PatchDataSave(Turbine.DataScope.Server, "DeedTracker_CharData", serverData);
end

function saveCurrentCharacterData()
    local charName = MYCHAR:GetName();
    if (SETTINGS.VERBOSE_OUTPUT) then
        Debug("Deed Tracker: Saving character data for " .. charName .. "...");
    end

    local mainWin = DeedTrackerWin.GetInstance();
    mainWin:SaveRegionExpandedStates();

    CompletionWindow.GetInstance():SaveUnclearedDeeds();

    local filename = GetCharDataFilename(charName);
    PatchDataSave(Turbine.DataScope.Server, filename, _CHARDATA[charName]);
end

function loadCharData(charName)
    local filename = GetCharDataFilename(charName);
    local savedCharData = PatchDataLoad(Turbine.DataScope.Server, filename);
    if (type(savedCharData) == 'table') then
        _CHARDATA[charName] = deepcopy(savedCharData);
    end
end

function loadData()
    local SavedSettings = {};
    SavedSettings = PatchDataLoad(Turbine.DataScope.Character, "DeedTracker_Settings");

    -- Check the saved settings to make sure it is still compatible with newer updates, add in any missing default settings
    if type(SavedSettings) == 'table' then
        local tempSETTINGS = {};
        tempSETTINGS = deepcopy(DEFAULT_SETTINGS);
        SETTINGS = mergeTables(tempSETTINGS,SavedSettings);
    else
        SETTINGS = deepcopy(DEFAULT_SETTINGS);
    end

    -- Load Server Data:
    local SavedCharData = {};
    SavedCharData = PatchDataLoad(Turbine.DataScope.Server, "DeedTracker_CharData");

    local isSaveFilePresent = false;
    -- Check the saved settings to make sure it is still compatible with newer updates, add in any missing default settings
    if type(SavedCharData) == 'table' then
        _CHARDATA = deepcopy(SavedCharData);
        isSaveFilePresent = true;
    end

    -- If this is V1 (all-in-one), update to V2 (per-character save file):
    if (_CHARDATA[":CHARACTERS"] == nil) then
        if (SETTINGS.VERBOSE_OUTPUT) then
            Debug("Deed Tracker: Upgrading save file format from v1 to v2");
        end

        -- Save out a backup in the original format if this isn't a brand new install:
        if (isSaveFilePresent) then
            if (SETTINGS.VERBOSE_OUTPUT) then
                Debug("Deed Tracker: Backing up save file to 'DeedTracker_CharData_v1backup'...");
            end
            PatchDataSave(Turbine.DataScope.Server, "DeedTracker_CharData_v1backup", _CHARDATA);
            if (SETTINGS.VERBOSE_OUTPUT) then
                Debug("  Deed Tracker: save file backup complete.");
            end
        end  

        -- Mark the version:
        _CHARDATA[":CHARDATA_VERSION"] = "v2.0";

        -- Store the character names:
        _CHARDATA[":CHARACTERS"] = GetAltNames();

        -- Write out a file for every existing character:
        for k,v in pairs (_CHARDATA) do
            if (k:sub(1, 1) ~= ":") then -- Don't do special categories
                if (SETTINGS.VERBOSE_OUTPUT) then
                    Debug("Deed Tracker: Creating individual save file for " .. k .. "...");
                end

                local filename = GetCharDataFilename(k);
                PatchDataSave(Turbine.DataScope.Server, filename, _CHARDATA[k]);

                if (SETTINGS.VERBOSE_OUTPUT) then
                    Debug("  Deed Tracker: individual save file created.");
                end                  
            end
        end

        -- Remove all char data from core file:
        saveServerSettings();

        -- v1 has been converted to v2, continue.
    else -- Already V2
        -- Load character data for all known characters:
        -- V1: This is already included in the above DeedTracker_CharData file
        -- V2: Iterate through each character in :CHARACTERS, load their file.
        for i=1, table.maxn(_CHARDATA[":CHARACTERS"]) do
            local name = _CHARDATA[":CHARACTERS"][i];
            loadCharData(name);
        end
    end


end

function updateData()
    -- v1.1.0:
    -- Realized we shouldn't store possible character names in _CHARDATA, so changed "SEARCH" and "SERVER" to ":SEARCH" and ":SERVER".
    -- Convert existing ones, to make like easier for testers, but remove eventually in case someone *does* have these names.
    if (_CHARDATA["SEARCH"] ~= nil) then
        _CHARDATA[":SEARCH"] = _CHARDATA["SEARCH"];
        _CHARDATA["SEARCH"] = nil;
    end

    if (_CHARDATA["SERVER"] ~= nil) then
        if (_CHARDATA["SERVER"]["DEEDS"] ~= nil) then
            _CHARDATA["SERVER"]["DEEDS"] = nil;
        end
        _CHARDATA[":SERVER"] = _CHARDATA["SERVER"];
        _CHARDATA["SERVER"] = nil;
    end

end

function DrawWindows()
    DeedTrackerWin.GetInstance():DrawMainWin();
    CompletionWindow.GetInstance():LoadUnclearedDeeds();
end

Turbine.Plugin.Unload = function (sender, args)
    -- Save the data when the plugin unloads.
    saveData();
    Debug(GetString(_LANG.STATUS.UNLOADED));
end

function SetServerLevelCap()
    local isLegendaryServer = LoadServerField("LEGENDARY_SERVER");

    local legendaryServerLevelCap = tonumber(LoadServerField("LEGENDARY_SERVER_LEVEL_CAP"));
    local serverLevelCap = tonumber(LoadServerField("SERVER_LEVEL_CAP"));

    if (LEGENDARY_SERVER_LEVEL_CAP > legendaryServerLevelCap) then
        if (SETTINGS.VERBOSE_OUTPUT) then
            Debug("Deed Tracker: Updating Legendary Server level cap to " .. LEGENDARY_SERVER_LEVEL_CAP);
        end

        legendaryServerLevelCap = LEGENDARY_SERVER_LEVEL_CAP;
        SaveServerField("LEGENDARY_SERVER_LEVEL_CAP", LEGENDARY_SERVER_LEVEL_CAP);
    end

    if (SERVER_LEVEL_CAP > serverLevelCap) then
        if (SETTINGS.VERBOSE_OUTPUT) then
            Debug("Deed Tracker: Updating Server level cap to " .. SERVER_LEVEL_CAP);
        end

        serverLevelCap = SERVER_LEVEL_CAP;
        SaveServerField("SERVER_LEVEL_CAP", SERVER_LEVEL_CAP);
    end

    if (isLegendaryServer) then
        CURRENT_LEVEL_CAP = legendaryServerLevelCap;
    else
        CURRENT_LEVEL_CAP = serverLevelCap;
    end
end

function RegisterForCombatChanged()
    MYCHAR.InCombatChanged = function() 
        CompletionWindow.GetInstance():InCombatChanged();
        DisambiguateDialog.GetInstance():InCombatChanged();
    end
end

function DetermineIfWaypointIsAvailable()
    -- Determine if Waypoint is available:
    local loadedPlugins = Turbine.PluginManager:GetAvailablePlugins();
    for i=1, table.maxn(loadedPlugins) do
        if (loadedPlugins[i].Name == "Waypoint") then
            WAYPOINT_AVAILABLE = true;
        end
    end

end

function Main()
    -- Make sure this isn't session play.
    local name = MYCHAR:GetName();
    if (name:sub(1,1) == "~") then
        Debug(GetString(_LANG.ERRORS.SESSION_PLAY));
        return;
    end

    -- Initiate load sequence
    UnloadReloader();
    loadData();
    updateData();
    CheckCharData();
    SetServerLevelCap();
    RegisterForCombatChanged();
    DrawWindows();
    InitiateChatLogger();
    RegisterCommands();
    CreateOptionsContent();
    DetermineIfWaypointIsAvailable();

    -- Load text
    Debug(GetString(_LANG.STATUS.LOADED));

    -- Check if session play completed a deed by delivering an item:
    CheckForDeedItems();
end

Main();
